<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Template\Engine;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\EntryPoint\ApplicationEntryPoint;
use Org\Mxchange\CoreFramework\Factory\ObjectFactory;
use Org\Mxchange\CoreFramework\Filesystem\FileNotFoundException;
use Org\Mxchange\CoreFramework\Manager\ManageableApplication;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;
use Org\Mxchange\CoreFramework\Registry\GenericRegistry;
use Org\Mxchange\CoreFramework\Response\Responseable;

// Import SPL stuff
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A generic template engine
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
abstract class BaseTemplateEngine extends BaseFrameworkSystem {
	/**
	 * The local path name where all templates and sub folders for special
	 * templates are stored. We will internally determine the language plus
	 * "html" for web templates or "emails" for email templates
	 */
	private $templateBasePath = '';

	/**
	 * Template type
	 */
	private $templateType = 'html';

	/**
	 * The extension for web and email templates (not compiled templates)
	 */
	private $templateExtension = '.tpl';

	/**
	 * The extension for code templates (not compiled templates)
	 */
	private $codeExtension = '.ctp';

	/**
	 * Path relative to $templateBasePath and language code for compiled code-templates
	 */
	private $compileOutputPath = 'templates/_compiled/';

	/**
	 * The path name for all templates
	 */
	private $genericBasePath = 'templates/';

	/**
	 * The raw (maybe uncompiled) template
	 */
	private $rawTemplateData = '';

	/**
	 * Template data with compiled-in variables
	 */
	private $compiledData = '';

	/**
	 * The last loaded template's file instance (SplFileInfo)
	 */
	private $lastTemplate = NULL;

	/**
	 * The variable stack for the templates
	 */
	private $varStack = array();

	/**
	 * Loaded templates for recursive protection and detection
	 */
	private $loadedTemplates = array();

	/**
	 * Compiled templates for recursive protection and detection
	 */
	private $compiledTemplates = array();

	/**
	 * Loaded raw template data
	 */
	private $loadedRawData = NULL;

	/**
	 * Raw templates which are linked in code templates
	 */
	private $rawTemplates = NULL;

	/**
	 * A regular expression for variable=value pairs
	 */
	private $regExpVarValue = '/([\w_]+)(="([^"]*)"|=([\w_]+))?/';

	/**
	 * A regular expression for filtering out code tags
	 *
	 * E.g.: {?template:variable=value;var2=value2;[...]?}
	 */
	private $regExpCodeTags = '/\{\?([a-z_]+)(:("[^"]+"|[^?}]+)+)?\?\}/';

	/**
	 * A regular expression to find template comments like <!-- Comment here //-->
	 */
	private $regExpComments = '/<!--[\w\W]*?(\/\/){0,1}-->/';

	/**
	 * Loaded helpers
	 */
	private $helpers = array();

	/**
	 * Current variable group
	 */
	private $currGroup = 'general';

	/**
	 * All template groups except "general"
	 */
	private $variableGroups = array();

	/**
	 * Code begin
	 */
	private $codeBegin = '<?php';

	/**
	 * Code end
	 */
	private $codeEnd = '?>';

	/**
	 * Language support is enabled by default
	 */
	private $languageSupport = true;

	/**
	 * XML compacting is disabled by default
	 */
	private $xmlCompacting = false;

	// Exception codes for the template engine
	const EXCEPTION_TEMPLATE_TYPE_IS_UNEXPECTED   = 0x110;
	const EXCEPTION_TEMPLATE_CONTAINS_INVALID_VAR = 0x111;
	const EXCEPTION_INVALID_VIEW_HELPER           = 0x112;
	const EXCEPTION_VARIABLE_IS_MISSING           = 0x113;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Init file I/O instance
		$ioInstance = ObjectFactory::createObjectByConfiguredName('file_io_class');

		// Set it
		$this->setFileIoInstance($ioInstance);
	}

	/**
	 * Search for a variable in the stack
	 *
	 * @param	$variableName	The variable we are looking for
	 * @param	$variableGroup	Optional variable group to look in
	 * @return	$index			false means not found, >=0 means found on a specific index
	 */
	private function getVariableIndex ($variableName, $variableGroup = NULL) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// First everything is not found
		$found = false;

		// If the stack is NULL, use the current group
		if (is_null($variableGroup)) {
			// Use current group
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.' currGroup=' . $this->currGroup . ' set as stack!');
			$variableGroup = $this->currGroup;
		} // END - if

		// Is the group there?
		if ($this->isVarStackSet($variableGroup)) {
			// Now search for it
			foreach ($this->getVarStack($variableGroup) as $index => $currEntry) {
				//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.':currGroup=' . $variableGroup . ',idx=' . $index . ',currEntry=' . $currEntry['name'] . ',variableName=' . $variableName);
				// Is the entry found?
				if ($currEntry['name'] == $variableName) {
					// Found!
					//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.':FOUND!');
					$found = $index;
					break;
				} // END - if
			} // END - foreach
		} // END - if

		// Return the current position
		return $found;
	}

	/**
	 * Checks whether the given variable group is set
	 *
	 * @param	$variableGroup	Variable group to check
	 * @return	$isSet			Whether the given variable group is set
	 */
	protected final function isVarStackSet ($variableGroup) {
		// Check it
		$isSet = isset($this->varStack[$variableGroup]);

		// Return result
		return $isSet;
	}

	/**
	 * Getter for given variable group
	 *
	 * @param	$variableGroup	Variable group to check
	 * @return	$varStack		Found variable group
	 */
	public final function getVarStack ($variableGroup) {
		return $this->varStack[$variableGroup];
	}

	/**
	 * Setter for given variable group
	 *
	 * @param	$variableGroup	Variable group to check
	 * @param	$varStack		Variable stack to check
	 * @return	void
	 */
	protected final function setVarStack ($variableGroup, array $varStack) {
		$this->varStack[$variableGroup]  = $varStack;
	}

	/**
	 * Return a content of a variable or null if not found
	 *
	 * @param	$variableName	The variable we are looking for
	 * @param	$variableGroup	Optional variable group to look in
	 * @return	$content		Content of the variable or null if not found
	 */
	protected function readVariable ($variableName, $variableGroup = NULL) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// First everything is not found
		$content = NULL;

		// If the stack is NULL, use the current group
		if (is_null($variableGroup)) {
			// Use current group
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.' currGroup=' . $this->currGroup . ' set as stack!');
			$variableGroup = $this->currGroup;
		} // END - if

		// Get variable index
		$found = $this->getVariableIndex($variableName, $variableGroup);

		// Is the variable found?
		if ($found !== false) {
			// Read it
			$content = $this->getVariableValue($variableGroup, $found);
		} // END - if

		// Return the current position
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.': variableGroup=' . $variableGroup . ',variableName=' . $variableName . ', content[' . gettype($content) . ']=' . $content);
		return $content;
	}

	/**
	 * Add a variable to the stack
	 *
	 * @param	$variableName	Name of variable to add
	 * @param	$value			Value we want to store in the variable
	 * @return	void
	 */
	private function addVariable ($variableName, $value) {
		// Set general variable group
		$this->setVariableGroup('general');

		// Add it to the stack
		$this->addGroupVariable($variableName, $value);
	}

	/**
	 * Returns all variables of current group or empty array
	 *
	 * @return	$result		Whether array of found variables or empty array
	 */
	private function readCurrentGroup () {
		// Default is not found
		$result = array();

		// Is the group there?
		if ($this->isVarStackSet($this->currGroup)) {
			// Then use it
			$result = $this->getVarStack($this->currGroup);
		} // END - if

		// Return result
		return $result;
	}

	/**
	 * Settter for variable group
	 *
	 * @param	$groupName	Name of variable group
	 * @param	$add		Whether add this group
	 * @return	void
	 */
	public function setVariableGroup ($groupName, $add = true) {
		// Set group name
		//* DEBIG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.': currGroup=' . $groupName);
		$this->currGroup = $groupName;

		// Skip group 'general'
		if (($groupName != 'general') && ($add === true)) {
			$this->variableGroups[$groupName] = 'OK';
		} // END - if
	}


	/**
	 * Adds a variable to current group
	 *
	 * @param	$variableName	Variable to set
	 * @param	$value			Value to store in variable
	 * @return	void
	 */
	public function addGroupVariable ($variableName, $value) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.': group=' . $this->currGroup . ', variableName=' . $variableName . ', value=' . $value);

		// Get current variables in group
		$currVars = $this->readCurrentGroup();

		// Append our variable
		array_push($currVars, $this->generateVariableArray($variableName, $value));

		// Add it to the stack
		$this->setVarStack($this->currGroup, $currVars);
	}

	/**
	 * Getter for variable value, throws a NoVariableException if the variable is not found
	 *
	 * @param	$variableGroup	Variable group to use
	 * @param	$index		Index in variable array
	 * @return	$value		Value to set
	 */
	private function getVariableValue ($variableGroup, $index) {
		// Return it
		return $this->varStack[$variableGroup][$index]['value'];
	}

	/**
	 * Modify an entry on the stack
	 *
	 * @param	$variableName	The variable we are looking for
	 * @param	$value			The value we want to store in the variable
	 * @return	void
	 * @throws	NoVariableException	If the given variable is not found
	 */
	private function modifyVariable ($variableName, $value) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Get index for variable
		$index = $this->getVariableIndex($variableName);

		// Is the variable set?
		if ($index === false) {
			// Unset variables cannot be modified
			throw new NoVariableException(array($this, $variableName, $value), self::EXCEPTION_VARIABLE_IS_MISSING);
		} // END - if

		// Then modify it
		$this->setVariableValue($this->currGroup, $index, $value);
	}

	/**
	 * Sets a variable value for given variable group and index
	 *
	 * @param	$variableGroup	Variable group to use
	 * @param	$index		Index in variable array
	 * @param	$value		Value to set
	 * @return	void
	 */
	private function setVariableValue ($variableGroup, $index, $value) {
		$this->varStack[$variableGroup][$index]['value'] = $value;
	}

	/**
	 * Sets a variable within given group. This method does detect if the
	 * variable is already set. If so, the variable got modified, otherwise
	 * added.
	 *
	 * @param	$variableGroup		Variable group to use
	 * @param	$variableName	Variable to set
	 * @param	$value			Value to set
	 * @return	void
	 */
	protected function setVariable ($variableGroup, $variableName, $value) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Get index for variable
		$index = $this->getVariableIndex($variableName);

		// Is the variable set?
		if ($index === false) {
			// Is the stack there?
			if (!isset($this->varStack[$variableGroup])) {
				// Then initialize it here
				$this->varStack[$variableGroup] = array();
			} // END - if

			// Not found, add it
			array_push($this->varStack[$variableGroup], $this->generateVariableArray($variableName, $value));
		} else {
			// Then modify it
			$this->setVariableValue($this->currGroup, $index, $value);
		}
	}

	/**
	 * "Generates" (better returns) an array for all variables for given
	 * variable/value pay.
	 *
	 * @param	$variableName	Variable to set
	 * @param	$value			Value to set
	 * @return	$varData		Variable data array
	 */
	private function generateVariableArray ($variableName, $value) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Generate the temporary array
		$varData = array(
			'name'  => $variableName,
			'value' => $value
		);

		// And return it
		return $varData;
	}

	/**
	 * Setter for template type. Only 'html', 'emails' and 'compiled' should
	 * be sent here
	 *
	 * @param	$templateType	The current template's type
	 * @return	void
	 */
	protected final function setTemplateType ($templateType) {
		$this->templateType = (string) $templateType;
	}

	/**
	 * Getter for template type
	 *
	 * @return	$templateType	The current template's type
	 */
	public final function getTemplateType () {
		return $this->templateType;
	}

	/**
	 * Setter for the last loaded template's file instance
	 *
	 * @param	$template	The last loaded template
	 * @return	void
	 */
	private final function setLastTemplate (SplFileInfo $fileInstance) {
		$this->lastTemplate = $fileInstance;
	}

	/**
	 * Getter for the last loaded template's file instance
	 *
	 * @return	$template	The last loaded template
	 */
	private final function getLastTemplate () {
		return $this->lastTemplate;
	}

	/**
	 * Setter for base path
	 *
	 * @param		$templateBasePath		The relative base path for all templates
	 * @return	void
	 */
	protected final function setTemplateBasePath ($templateBasePath) {
		// And set it
		$this->templateBasePath = (string) $templateBasePath;
	}

	/**
	 * Getter for base path
	 *
	 * @return	$templateBasePath		The relative base path for all templates
	 */
	public final function getTemplateBasePath () {
		// And set it
		return $this->templateBasePath;
	}

	/**
	 * Getter for generic base path
	 *
	 * @return	$templateBasePath		The relative base path for all templates
	 */
	public final function getGenericBasePath () {
		// And set it
		return $this->genericBasePath;
	}

	/**
	 * Setter for template extension
	 *
	 * @param		$templateExtension	The file extension for all uncompiled
	 *							templates
	 * @return	void
	 */
	protected final function setRawTemplateExtension ($templateExtension) {
		// And set it
		$this->templateExtension = (string) $templateExtension;
	}

	/**
	 * Setter for code template extension
	 *
	 * @param		$codeExtension		The file extension for all uncompiled
	 *							templates
	 * @return	void
	 */
	protected final function setCodeTemplateExtension ($codeExtension) {
		// And set it
		$this->codeExtension = (string) $codeExtension;
	}

	/**
	 * Getter for template extension
	 *
	 * @return	$templateExtension	The file extension for all uncompiled
	 *							templates
	 */
	public final function getRawTemplateExtension () {
		// And set it
		return $this->templateExtension;
	}

	/**
	 * Getter for code-template extension
	 *
	 * @return	$codeExtension		The file extension for all code-
	 *							templates
	 */
	public final function getCodeTemplateExtension () {
		// And set it
		return $this->codeExtension;
	}

	/**
	 * Setter for path of compiled templates
	 *
	 * @param	$compileOutputPath	The local base path for all compiled
	 *								templates
	 * @return	void
	 */
	protected final function setCompileOutputPath ($compileOutputPath) {
		// And set it
		$this->compileOutputPath = (string) $compileOutputPath;
	}

	/**
	 * Unsets the given offset in the variable group
	 *
	 * @param	$index			Index to unset
	 * @param	$variableGroup	Variable group (default: currGroup)
	 * @return	void
	 */
	protected final function unsetVariableStackOffset ($index, $variableGroup = NULL) {
		// Is the variable group not set?
		if (is_null($variableGroup)) {
			// Then set it to current
			$variableGroup = $this->currGroup;
		} // END - if

		// Is the entry there?
		if (!isset($this->varStack[$variableGroup][$index])) {
			// Abort here, we need fixing!
			$this->debugInstance();
		} // END - if

		// Remove it
		unset($this->varStack[$variableGroup][$index]);
	}

	/**
	 * Private setter for raw template data
	 *
	 * @param	$rawTemplateData	The raw data from the template
	 * @return	void
	 */
	protected final function setRawTemplateData ($rawTemplateData) {
		// And store it in this class
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.': ' . strlen($rawTemplateData) . ' Bytes set.');
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(__METHOD__.': ' . $this->currGroup . ' variables: ' . count($this->getVarStack($this->currGroup)) . ', groups=' . count($this->varStack));
		$this->rawTemplateData = (string) $rawTemplateData;
	}

	/**
	 * Getter for raw template data
	 *
	 * @return	$rawTemplateData	The raw data from the template
	 */
	public final function getRawTemplateData () {
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: ' . strlen($this->rawTemplateData) . ' Bytes read.');
		return $this->rawTemplateData;
	}

	/**
	 * Private setter for compiled templates
	 *
	 * @return	void
	 */
	private final function setCompiledData ($compiledData) {
		// And store it in this class
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: ' . strlen($compiledData) . ' Bytes set.');
		$this->compiledData = (string) $compiledData;
	}

	/**
	 * Getter for compiled templates, must be public for e.g. Mailer classes.
	 *
	 * @return	$compiledData	Compiled template data
	 */
	public final function getCompiledData () {
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: ' . strlen($this->compiledData) . ' Bytes read.');
		return $this->compiledData;
	}

	/**
	 * Private loader for all template types
	 *
	 * @param	$templateName	The template we shall load
	 * @param	$extOther	An other extension to use
	 * @return	void
	 * @throws	FileNotFoundException	If the template was not found
	 */
	protected function loadTemplate ($templateName, $extOther = '') {
		// Get extension for the template if empty
		if (empty($extOther)) {
			// None provided, so get the raw one
			$ext = $this->getRawTemplateExtension();
		} else {
			// Then use it!
			$ext = (string) $extOther;
		}

		/*
		 * Construct the FQFN for the template without language as language is
		 * now entirely done by php_intl. These old thing with language-based
		 * template paths comes from an older time.
		 */
		$fileInstance = new SplFileInfo(sprintf('%s%s%s%s%s%s',
			$this->getTemplateBasePath(),
			$this->getGenericBasePath(),
			$this->getTemplateType(),
			DIRECTORY_SEPARATOR,
			(string) $templateName,
			$ext
		));

		// First try this
		try {
			// Load the raw template data
			$this->loadRawTemplateData($fileInstance);
		} catch (FileNotFoundException $e) {
			// If we shall load a code-template we need to switch the file extension
			if (($this->getTemplateType() != $this->getConfigInstance()->getConfigEntry('html_template_type')) && (empty($extOther))) {
				// Switch over to the code-template extension and try it again
				$ext = $this->getCodeTemplateExtension();

				// Try it again...
				$this->loadTemplate($templateName, $ext);
			} else {
				// Throw it again
				throw new FileNotFoundException($fileInstance, self::EXCEPTION_FILE_NOT_FOUND);
			}
		}

	}

	/**
	 * A private loader for raw template names
	 *
	 * @param	$fileInstance	An instance of a SplFileInfo class
	 * @return	void
	 */
	private function loadRawTemplateData (SplFileInfo $fileInstance) {
		// Some debug code to look on the file which is being loaded
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: fileInstance=' . $fileInstance);

		// Load the raw template
		$rawTemplateData = $this->getFileIoInstance()->loadFileContents($fileInstance);

		// Store the template's contents into this class
		$this->setRawTemplateData($rawTemplateData);

		// Remember the template's file instance
		$this->setLastTemplate($fileInstance);
	}

	/**
	 * Try to assign an extracted template variable as a "content" or 'config'
	 * variable.
	 *
	 * @param	$variableName		The variable's name (shall be content or config)
	 *							by default
	 * @param	$variableName	The variable we want to assign
	 * @return	void
	 */
	private function assignTemplateVariable ($variableName, $var) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: variableName=' . $variableName . ',variableName=' . $variableName);

		// Is it not a config variable?
		if ($variableName != 'config') {
			// Regular template variables
			$this->assignVariable($variableName, '');
		} else {
			// Configuration variables
			$this->assignConfigVariable($var);
		}
	}

	/**
	 * Extract variables from a given raw data stream
	 *
	 * @param	$rawData	The raw template data we shall analyze
	 * @return	void
	 */
	private function extractVariablesFromRawData ($rawData) {
		// Cast to string
		$rawData = (string) $rawData;

		// Search for variables
		preg_match_all('/\$(\w+)(\[(\w+)\])?/', $rawData, $variableMatches);

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:rawData(' . strlen($rawData) . ')=' . $rawData . ',variableMatches=' . print_r($variableMatches, true));

		// Did we find some variables?
		if ((is_array($variableMatches)) && (count($variableMatches) == 4) && (count($variableMatches[0]) > 0)) {
			// Initialize all missing variables
			foreach ($variableMatches[3] as $key => $var) {
				// Variable name
				$variableName = $variableMatches[1][$key];

				// Workarround: Do not assign empty variables
				if (!empty($var)) {
					// Try to assign it, empty strings are being ignored
					$this->assignTemplateVariable($variableName, $var);
				} // END - if
			} // END - foreach
		} // END - if
	}

	/**
	 * Main analysis of the loaded template
	 *
	 * @param	$templateMatches	Found template place-holders, see below
	 * @return	void
	 *
	 *---------------------------------
	 * Structure of $templateMatches:
	 *---------------------------------
	 * [0] => Array - An array with all full matches
	 * [1] => Array - An array with left part (before the ':') of a match
	 * [2] => Array - An array with right part of a match including ':'
	 * [3] => Array - An array with right part of a match excluding ':'
	 */
	private function analyzeTemplate (array $templateMatches) {
		// Backup raw template data
		$backup = $this->getRawTemplateData();

		// Initialize some arrays
		if (is_null($this->loadedRawData)) {
			// Initialize both
			$this->loadedRawData = array();
			$this->rawTemplates = array();
		} // END - if

		// Load all requested templates
		foreach ($templateMatches[1] as $template) {
			// Load and compile only templates which we have not yet loaded
			// RECURSIVE PROTECTION! BE CAREFUL HERE!
			if ((!isset($this->loadedRawData[$template])) && (!in_array($template, $this->loadedTemplates))) {
				// Debug message
				//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:template=' . $template);

				// Template not found, but maybe variable assigned?
				if ($this->getVariableIndex($template) !== false) {
					// Use that content here
					$this->loadedRawData[$template] = $this->readVariable($template);

					// Recursive protection:
					array_push($this->loadedTemplates, $template);
				} else {
					// Then try to search for code-templates
					try {
						// Load the code template and remember it's contents
						$this->loadCodeTemplate($template);
						$this->loadedRawData[$template] = $this->getRawTemplateData();

						// Remember this template for recursion detection
						// RECURSIVE PROTECTION!
						array_push($this->loadedTemplates, $template);
					} catch (FileNotFoundException $e) {
						// Even this is not done... :/
						array_push($this->rawTemplates, $template);
					}
				}
			} // END - if
		} // END - foreach

		// Restore the raw template data
		$this->setRawTemplateData($backup);
	}

	/**
	 * Compile a given raw template code and remember it for later usage
	 *
	 * @param	$code		The raw template code
	 * @param	$template	The template's name
	 * @return	void
	 */
	private function compileCode ($code, $template) {
		// Is this template already compiled?
		if (in_array($template, $this->compiledTemplates)) {
			// Abort here...
			return;
		} // END - if

		// Remember this template being compiled
		array_push($this->compiledTemplates, $template);

		// Compile the loaded code in five steps:
		//
		// 1. Backup current template data
		$backup = $this->getRawTemplateData();

		// 2. Set the current template's raw data as the new content
		$this->setRawTemplateData($code);

		// 3. Compile the template data
		$this->compileTemplate();

		// 4. Remember it's contents
		$this->loadedRawData[$template] = $this->getRawTemplateData();

		// 5. Restore the previous raw content from backup variable
		$this->setRawTemplateData($backup);
	}

	/**
	 * Insert all given and loaded templates by running through all loaded
	 * codes and searching for their place-holder in the main template
	 *
	 * @param	$templateMatches	See method analyzeTemplate()
	 * @return	void
	 */
	private function insertAllTemplates (array $templateMatches) {
		// Run through all loaded codes
		foreach ($this->loadedRawData as $template => $code) {

			// Search for the template
			$foundIndex = array_search($template, $templateMatches[1]);

			// Lookup the matching template replacement
			if (($foundIndex !== false) && (isset($templateMatches[0][$foundIndex]))) {

				// Get the current raw template
				$rawData = $this->getRawTemplateData();

				// Replace the space holder with the template code
				$rawData = str_replace($templateMatches[0][$foundIndex], $code, $rawData);

				// Set the new raw data
				$this->setRawTemplateData($rawData);
			} // END - if
		} // END - foreach
	}

	/**
	 * Load all extra raw templates
	 *
	 * @return	void
	 */
	private function loadExtraRawTemplates () {
		// Are there some raw templates we need to load?
		if (count($this->rawTemplates) > 0) {
			// Try to load all raw templates
			foreach ($this->rawTemplates as $key => $template) {
				try {
					// Load the template
					$this->loadHtmlTemplate($template);

					// Remember it's contents
					$this->rawTemplates[$template] = $this->getRawTemplateData();

					// Remove it from the loader list
					unset($this->rawTemplates[$key]);

					// Remember this template for recursion detection
					// RECURSIVE PROTECTION!
					array_push($this->loadedTemplates, $template);
				} catch (FileNotFoundException $e) {
					// This template was never found. We silently ignore it
					unset($this->rawTemplates[$key]);
				}
			} // END - foreach
		} // END - if
	}

	/**
	 * Assign all found template variables
	 *
	 * @param	$varMatches		An array full of variable/value pairs.
	 * @return	void
	 * @todo	Unfinished work or don't die here.
	 */
	private function assignAllVariables (array $varMatches) {
		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:varMatches()=' . count($varMatches));

		// Search for all variables
		foreach ($varMatches[1] as $key => $var) {
			// Replace all dashes to underscores to match variables with configuration entries
			$var = trim(self::convertDashesToUnderscores($var));

			// Debug message
			self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:key=' . $key . ',var=' . $var);

			// Detect leading equals
			if (substr($varMatches[2][$key], 0, 1) == '=') {
				// Remove and cast it
				$varMatches[2][$key] = (string) substr($varMatches[2][$key], 1);
			} // END - if

			// Do we have some quotes left and right side? Then it is free text
			if ((substr($varMatches[2][$key], 0, 1) == '"') && (substr($varMatches[2][$key], -1, 1) == '"')) {
				// Free string detected! Which we can assign directly
				$this->assignVariable($var, $varMatches[3][$key]);
			} elseif (!empty($varMatches[2][$key])) {
				// @TODO Non-string found so we need some deeper analysis...
				ApplicationEntryPoint::exitApplication('Deeper analysis not yet implemented!');
			}
		} // END - foreach
	}

	/**
	 * Compiles all loaded raw templates
	 *
	 * @param	$templateMatches	See method analyzeTemplate() for details
	 * @return	void
	 */
	private function compileRawTemplateData (array $templateMatches) {
		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:loadedRawData()= ' .count($this->loadedRawData));

		// Are some code-templates found which we need to compile?
		if (count($this->loadedRawData) > 0) {
			// Then compile all!
			foreach ($this->loadedRawData as $template => $code) {
				// Debug message
				//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:template=' . $template . ',code(' . strlen($code) . ')=' . $code);

				// Is this template already compiled?
				if (in_array($template, $this->compiledTemplates)) {
					// Then skip it
					//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: Template ' . $template . ' already compiled. SKIPPED!');
					continue;
				} // END - if

				// Search for the template
				$foundIndex = array_search($template, $templateMatches[1]);

				// Lookup the matching variable data
				if (($foundIndex !== false) && (isset($templateMatches[3][$foundIndex]))) {
					// Split it up with another reg. exp. into variable=value pairs
					preg_match_all($this->regExpVarValue, $templateMatches[3][$foundIndex], $varMatches);
					//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:varMatches=' . print_r($varMatches, true));

					// Assign all variables
					$this->assignAllVariables($varMatches);
				} // END - if (isset($templateMatches ...

				// Compile the loaded template
				$this->compileCode($code, $template);
			} // END - foreach ($this->loadedRawData ...

			// Insert all templates
			$this->insertAllTemplates($templateMatches);
		} // END - if (count($this->loadedRawData) ...
	}

	/**
	 * Inserts all raw templates into their respective variables
	 *
	 * @return	void
	 */
	private function insertRawTemplates () {
		// Load all templates
		foreach ($this->rawTemplates as $template => $content) {
			// Set the template as a variable with the content
			$this->assignVariable($template, $content);
		}
	}

	/**
	 * Finalizes the compilation of all template variables
	 *
	 * @return	void
	 */
	private function finalizeVariableCompilation () {
		// Get the content
		$content = $this->getRawTemplateData();
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: content before=' . strlen($content) . ' (' . md5($content) . ')');

		// Do we have the stack?
		if (!$this->isVarStackSet('general')) {
			// Abort here silently
			// @TODO This silent abort should be logged, maybe.
			return;
		} // END - if

		// Walk through all variables
		foreach ($this->getVarStack('general') as $currEntry) {
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: name=' . $currEntry['name'] . ', value=<pre>' . htmlentities($currEntry['value']) . '</pre>');
			// Replace all [$var] or {?$var?} with the content
			// @TODO Old behaviour, will become obsolete!
			$content = str_replace('$content[' . $currEntry['name'] . ']', $currEntry['value'], $content);

			// @TODO Yet another old way
			$content = str_replace('[' . $currEntry['name'] . ']', $currEntry['value'], $content);

			// The new behaviour
			$content = str_replace('{?' . $currEntry['name'] . '?}', $currEntry['value'], $content);
		} // END - for

		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: content after=' . strlen($content) . ' (' . md5($content) . ')');

		// Set the content back
		$this->setRawTemplateData($content);
	}

	/**
	 * Load a specified HTML template into the engine
	 *
	 * @param	$template	The web template we shall load which is located in
	 *						'html' by default
	 * @return	void
	 */
	public function loadHtmlTemplate ($template) {
		// Set template type
		$this->setTemplateType($this->getConfigInstance()->getConfigEntry('html_template_type'));

		// Load the special template
		$this->loadTemplate($template);
	}

	/**
	 * Assign (add) a given variable with a value
	 *
	 * @param	$variableName	The variable we are looking for
	 * @param	$value			The value we want to store in the variable
	 * @return	void
	 * @throws	InvalidArgumentException	If the variable name is left empty
	 */
	public final function assignVariable ($variableName, $value) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Empty variable found?
		if (empty($variableName)) {
			// Throw an exception
			throw new InvalidArgumentException('Parameter "variableName" is empty');
		} // END - if

		// First search for the variable if it was already added
		$index = $this->getVariableIndex($variableName);

		// Was it found?
		if ($index === false) {
			// Add it to the stack
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:ADD: ' . $variableName . '[' . gettype($value) . ']=' . $value);
			$this->addVariable($variableName, $value);
		} elseif (!empty($value)) {
			// Modify the stack entry
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:MOD: ' . $variableName . '[' . gettype($value) . ']=' . $value);
			$this->modifyVariable($variableName, $value);
		}
	}

	/**
	 * Removes a given variable
	 *
	 * @param	$variableName	The variable we are looking for
	 * @param	$variableGroup	Name of variable group (default: 'general')
	 * @return	void
	 */
	public final function removeVariable ($variableName, $variableGroup = 'general') {
		// First search for the variable if it was already added
		$index = $this->getVariableIndex($variableName, $variableGroup);

		// Was it found?
		if ($index !== false) {
			// Remove this variable
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:UNSET: variableGroup=' . $variableGroup . ',variableName=' . $variableName . ',index=' . $index);
			$this->unsetVariableStackOffset($index, $variableGroup);
		} // END - if
	}

	/**
	 * Assigns the last loaded raw template content with a given variable
	 *
	 * @param	$templateName	Name of the template we want to assign
	 * @param	$variableName	Name of the variable we want to assign
	 * @return	void
	 */
	public function assignTemplateWithVariable ($templateName, $variableName) {
		// Get the content from last loaded raw template
		$content = $this->getRawTemplateData();

		// Assign the variable
		$this->assignVariable($variableName, $content);

		// Purge raw content
		$this->setRawTemplateData('');
	}

	/**
	 * Assign a given congfiguration variable with a value
	 *
	 * @param	$variableName	The configuration variable we want to assign
	 * @return	void
	 */
	public function assignConfigVariable ($variableName) {
		// Replace all dashes to underscores to match variables with configuration entries
		$variableName = trim(self::convertDashesToUnderscores($variableName));

		// Sweet and simple...
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: variableName=' . $variableName . ',getConfigEntry()=' . $this->getConfigInstance()->getConfigEntry($variableName));
		$this->assignVariable($variableName, $this->getConfigInstance()->getConfigEntry($variableName));
	}

	/**
	 * Assigns a lot variables into the stack of currently loaded template.
	 * This method should only be used in very rare circumstances, e.g. when
	 * you have to copy a whole set of variables into the template engine.
	 * Before you use this method, please make sure you have considered all
	 * other possiblities.
	 *
	 * @param	$variables	An array with variables to be assigned
	 * @return	void
	 */
	public function assignMultipleVariables (array $variables) {
		// "Inject" all
		foreach ($variables as $name => $value) {
			// Set variable with name for 'config' group
			$this->assignVariable($name, $value);
		} // END - foreach
	}

	/**
	 * Assigns all the application data with template variables
	 *
	 * @return	void
	 */
	public function assignApplicationData () {
		// Get application instance
		$applicationInstance = GenericRegistry::getRegistry()->getInstance('application');

		// Get long name and assign it
		$this->assignVariable('app_full_name' , $applicationInstance->getAppName());

		// Get short name and assign it
		$this->assignVariable('app_short_name', $applicationInstance->getAppShortName());

		// Get version number and assign it
		$this->assignVariable('app_version'   , $applicationInstance->getAppVersion());

		// Assign extra application-depending data
		$applicationInstance->assignExtraTemplateData($this);
	}

	/**
	 * Load a specified code template into the engine
	 *
	 * @param	$template	The code template we shall load which is
	 *						located in 'code' by default
	 * @return	void
	 */
	public function loadCodeTemplate ($template) {
		// Set template type
		$this->setTemplateType($this->getConfigInstance()->getConfigEntry('code_' . FrameworkBootstrap::getRequestTypeFromSystem() . '_template_type'));

		// Load the special template
		$this->loadTemplate($template);
	}

	/**
	 * Load a specified email template into the engine
	 *
	 * @param	$template	The email template we shall load which is
	 *						located in 'emails' by default
	 * @return	void
	 */
	public function loadEmailTemplate ($template) {
		// Set template type
		$this->setTemplateType($this->getConfigInstance()->getConfigEntry('email_template_type'));

		// Load the special template
		$this->loadTemplate($template);
	}

	/**
	 * Compiles configuration place-holders in all variables. This 'walks'
	 * through the variable group 'general'. It interprets all values from that
	 * variables as configuration entries after compiling them.
	 *
	 * @return	void
	 */
	public final function compileConfigInVariables () {
		// Do we have the stack?
		if (!$this->isVarStackSet('general')) {
			// Abort here silently
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: Aborted, variable stack general not found!');
			return;
		} // END - if

		// Iterate through all general variables
		foreach ($this->getVarStack('general') as $index => $currVariable) {
			// Compile the value
			$value = $this->compileRawCode($this->readVariable($currVariable['name']), true);

			// Debug message
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: name=' . $currVariable['name'] . ',value=' . $value);

			// Remove it from stack
			$this->removeVariable($currVariable['name'], 'general');
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: value='. $value . ',name=' . $currVariable['name'] . ',index=' . $index);

			// Is it a configuration key?
			if ($this->getConfigInstance()->isConfigurationEntrySet($value)) {
				// The value itself is a configuration entry
				$this->assignConfigVariable($value);
			} else {
				// Re-assign the value directly
				$this->assignVariable($currVariable['name'], $value);
			}
		} // END - foreach
	}

	/**
	 * Compile all variables by inserting their respective values
	 *
	 * @return	void
	 * @todo	Make this code some nicer...
	 */
	public final function compileVariables () {
		// Initialize the $content array
		$validVar = $this->getConfigInstance()->getConfigEntry('tpl_valid_var');
		$dummy = array();

		// Iterate through all general variables
		foreach ($this->getVarStack('general') as $currVariable) {
			// Transfer it's name/value combination to the $content array
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:' . $currVariable['name'] . '=<pre>' . htmlentities($currVariable['value']).'</pre>');
			$dummy[$currVariable['name']] = $currVariable['value'];
		}// END - if

		// Set the new variable (don't remove the second dollar!)
		$$validVar = $dummy;

		// Remove some variables
		unset($index);
		unset($currVariable);

		// Run the compilation three times to get content from helper classes in
		$cnt = 0;
		while ($cnt < 3) {
			// Finalize the compilation of template variables
			$this->finalizeVariableCompilation();

			// Prepare the eval() command for comiling the template
			$eval = sprintf('$result = "%s";',
				addslashes($this->getRawTemplateData())
			);

			// This loop does remove the backslashes (\) in PHP parameters
			while (strpos($eval, $this->codeBegin) !== false) {
				// Get left part before "<?"
				$evalLeft = substr($eval, 0, strpos($eval, $this->codeBegin));

				// Get all from right of "<?"
				$evalRight = substr($eval, (strpos($eval, $this->codeBegin) + 5));

				// Cut middle part out and remove escapes
				$evalMiddle = trim(substr($evalRight, 0, strpos($evalRight, $this->codeEnd)));
				$evalMiddle = stripslashes($evalMiddle);

				// Remove the middle part from right one
				$evalRight = substr($evalRight, (strpos($evalRight, $this->codeEnd) + 2));

				// And put all together
				$eval = sprintf('%s<%%php %s %%>%s', $evalLeft, $evalMiddle, $evalRight);
			} // END - while

			// Prepare PHP code for eval() command
			$eval = str_replace(
				'<%php', '";',
				str_replace(
					'%>',
					"\n\$result .= \"",
					$eval
				)
			);

			// Run the constructed command. This will "compile" all variables in
			eval($eval);

			// Goes something wrong?
			if ((!isset($result)) || (empty($result))) {
				// Output eval command
				self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('Failed eval() code: <pre>%s</pre>', $this->markupCode($eval, true)), true);

				// Output backtrace here
				$this->debugBackTrace();
			} // END - if

			// Set raw template data
			$this->setRawTemplateData($result);
			$cnt++;
		} // END - while

		// Final variable assignment
		$this->finalizeVariableCompilation();

		// Set the new content
		$this->setCompiledData($this->getRawTemplateData());
	}

	/**
	 * Compile all required templates into the current loaded one
	 *
	 * @return	void
	 * @throws	UnexpectedTemplateTypeException	If the template type is
	 *											not "code"
	 * @throws	InvalidArrayCountException		If an unexpected array
	 *											count has been found
	 */
	public function compileTemplate () {
		// Get code type to make things shorter
		$codeType = $this->getConfigInstance()->getConfigEntry('code_' . FrameworkBootstrap::getRequestTypeFromSystem() . '_template_type');

		// We will only work with template type "code" from configuration
		if (substr($this->getTemplateType(), 0, strlen($codeType)) != $codeType) {
			// Abort here
			throw new UnexpectedTemplateTypeException(array($this, $this->getTemplateType(), $this->getConfigInstance()->getConfigEntry('code_' . FrameworkBootstrap::getRequestTypeFromSystem() . '_template_type')), self::EXCEPTION_TEMPLATE_TYPE_IS_UNEXPECTED);
		} // END - if

		// Get the raw data.
		$rawData = $this->getRawTemplateData();

		// Remove double spaces and trim leading/trailing spaces
		$rawData = trim(str_replace('  ', ' ', $rawData));

		// Search for raw variables
		$this->extractVariablesFromRawData($rawData);

		// Search for code-tags which are {? ?}
		preg_match_all($this->regExpCodeTags, $rawData, $templateMatches);

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:templateMatches=' . print_r($templateMatches , true));

		// Analyze the matches array
		if ((is_array($templateMatches)) && (count($templateMatches) == 4) && (count($templateMatches[0]) > 0)) {
			// Entries are found:
			//
			// The main analysis
			$this->analyzeTemplate($templateMatches);

			// Compile raw template data
			$this->compileRawTemplateData($templateMatches);

			// Are there some raw templates left for loading?
			$this->loadExtraRawTemplates();

			// Are some raw templates found and loaded?
			if (count($this->rawTemplates) > 0) {
				// Insert all raw templates
				$this->insertRawTemplates();

				// Remove the raw template content as well
				$this->setRawTemplateData('');
			} // END - if
		} // END - if($templateMatches ...
	}

	/**
	 * Loads a given view helper (by name)
	 *
	 * @param	$helperName		The helper's name
	 * @return	void
	 */
	protected function loadViewHelper ($helperName) {
		// Is this view helper loaded?
		if (!isset($this->helpers[$helperName])) {
			// Create a class name
			$className = self::convertToClassName($helperName) . 'ViewHelper';

			// Generate new instance
			$this->helpers[$helperName] = ObjectFactory::createObjectByName($className);
		} // END - if

		// Return the requested instance
		return $this->helpers[$helperName];
	}

	/**
	 * Transfers the content of this template engine to a given response instance
	 *
	 * @param	$responseInstance	An instance of a Responseable class
	 * @return	void
	 */
	public function transferToResponse (Responseable $responseInstance) {
		// Get the content and set it in response class
		$responseInstance->writeToBody($this->getCompiledData());
	}

	/**
	 * "Compiles" a variable by replacing {?var?} with it's content
	 *
	 * @param	$rawCode			Raw code to compile
	 * @param	$setMatchAsCode		Sets $match if readVariable() returns empty result
	 * @return	$rawCode	Compile code with inserted variable value
	 */
	public function compileRawCode ($rawCode, $setMatchAsCode=false) {
		// Find the variables
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:rawCode=<pre>' . htmlentities($rawCode) . '</pre>');
		preg_match_all($this->regExpVarValue, $rawCode, $varMatches);

		// Compile all variables
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:<pre>' . print_r($varMatches, true) . '</pre>');
		foreach ($varMatches[0] as $match) {
			// Add variable tags around it
			$varCode = '{?' . $match . '?}';

			// Debug message
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:varCode=' . $varCode);

			// Is the variable found in code? (safes some calls)
			if (strpos($rawCode, $varCode) !== false) {
				// Debug message
				//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: match=' . $match . ',rawCode[' . gettype($rawCode) . ']=' . $rawCode);

				// Use $match as new value or $value from read variable?
				if ($setMatchAsCode === true) {
					// Insert match
					$rawCode = str_replace($varCode, $match, $rawCode);
				} else {
					// Read the variable
					$value = $this->readVariable($match);

					// Insert value
					$rawCode = str_replace($varCode, $value, $rawCode);
				}
			} // END - if
		} // END - foreach

		// Return the compiled data
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']:rawCode=<pre>' . htmlentities($rawCode) . '</pre>');
		return $rawCode;
	}

	/**
	 * Getter for variable group array
	 *
	 * @return	$variableGroups	All variable groups
	 */
	public final function getVariableGroups () {
		return $this->variableGroups;
	}

	/**
	 * Renames a variable in code and in stack
	 *
	 * @param	$oldName	Old name of variable
	 * @param	$newName	New name of variable
	 * @return	void
	 */
	public function renameVariable ($oldName, $newName) {
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('BASE-TEMPLATE[' . __METHOD__ . ':' . __LINE__ . ']: oldName=' . $oldName . ', newName=' . $newName);
		// Get raw template code
		$rawData = $this->getRawTemplateData();

		// Replace it
		$rawData = str_replace($oldName, $newName, $rawData);

		// Set the code back
		$this->setRawTemplateData($rawData);
	}

	/**
	 * Renders the given XML content
	 *
	 * @param	$content	Valid XML content or if not set the current loaded raw content
	 * @return	void
	 * @throws	XmlParserException	If an XML error was found
	 */
	public function renderXmlContent ($content = NULL) {
		// Is the content set?
		if (is_null($content)) {
			// Get current content
			$content = $this->getRawTemplateData();
		} // END - if

		// Get a XmlParser instance
		$parserInstance = ObjectFactory::createObjectByConfiguredName('xml_parser_class', array($this));

		// Check if XML compacting is enabled
		if ($this->isXmlCompactingEnabled()) {
			// Yes, so get a decorator class for transparent compacting
			$parserInstance = ObjectFactory::createObjectByConfiguredName('deco_compacting_xml_parser_class', array($parserInstance));
		} // END - if

		// Parse the XML document
		$parserInstance->parseXmlContent($content);
	}

	/**
	 * Enables or disables language support
	 *
	 * @param	$languageSupport	New language support setting
	 * @return	void
	 */
	public final function enableLanguageSupport ($languageSupport = true) {
		$this->languageSupport = (bool) $languageSupport;
	}

	/**
	 * Checks whether language support is enabled
	 *
	 * @return	$languageSupport	Whether language support is enabled or disabled
	 */
	public final function isLanguageSupportEnabled () {
		return $this->languageSupport;
	}

	/**
	 * Enables or disables XML compacting
	 *
	 * @param	$xmlCompacting	New XML compacting setting
	 * @return	void
	 */
	public final function enableXmlCompacting ($xmlCompacting = true) {
		$this->xmlCompacting = (bool) $xmlCompacting;
	}

	/**
	 * Checks whether XML compacting is enabled
	 *
	 * @return	$xmlCompacting	Whether XML compacting is enabled or disabled
	 */
	public final function isXmlCompactingEnabled () {
		return $this->xmlCompacting;
	}

	/**
	 * Removes all commentd, tabs and new-line characters to compact the content
	 *
	 * @param	$uncompactedContent		The uncompacted content
	 * @return	$compactedContent		The compacted content
	 */
	public function compactContent ($uncompactedContent) {
		// First, remove all tab/new-line/revert characters
		$compactedContent = str_replace(chr(9), '', str_replace(chr(10), '', str_replace(chr(13), '', $uncompactedContent)));

		// Then regex all comments like <!-- //--> away
		preg_match_all($this->regExpComments, $compactedContent, $matches);

		// Do we have entries?
		if (isset($matches[0][0])) {
			// Remove all
			foreach ($matches[0] as $match) {
				// Remove the match
				$compactedContent = str_replace($match, '', $compactedContent);
			} // END - foreach
		} // END - if

		// Set the content again
		$this->setRawTemplateData($compactedContent);

		// Return compacted content
		return $compactedContent;
	}

}
